/*
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 * 
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *          http://www.eclipse.org/legal/epl-v10.html
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.field.type;

import com.vividsolutions.jts.geom.Geometry;
import java.awt.Image;
import java.awt.datatransfer.Transferable;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import javax.swing.Action;
import nl.cloudfarming.client.field.model.ActivityField;
import nl.cloudfarming.client.field.model.Shape;
import nl.cloudfarming.client.field.model.ShapeCollection;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import nl.cloudfarming.client.util.actions.OpenAction;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.nodes.NodeTransfer;
import org.openide.util.ImageUtilities;
import org.openide.util.Utilities;
import org.openide.util.datatransfer.PasteType;
import org.openide.util.lookup.Lookups;

/**
 * A node representing {@link ActivityField} object
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
public final class ActivityFieldNode extends AbstractNode implements Serializable {

    private static final Image icon = ImageUtilities.loadImage("nl/cloudfarming/client/field/type/folder_treatmentzones.png", true);
    private ActivityField activityField;

    /**
     * Constructs a ActivityFieldNode
     *
     * @param activityField the {@link ActivityField} for which this node is
     */
    public ActivityFieldNode(ActivityField activityField) {
        super(Children.create(new ActivityFieldChildFactory(activityField), true),
                Lookups.fixed(activityField, new SimpleLayer(activityField.getName())));
        this.activityField = activityField;
        setName(activityField.getName());
        activityField.addPropertyChangeListener(new PropertyChangeListener() {
            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                if (ActivityField.PROP_NAME.equals(evt.getPropertyName())) {
                    setName((String) evt.getNewValue());
                }
            }
        });
    }

    @Override
    public Action[] getActions(boolean context) {
        List<Action> actions = new ArrayList<>();
        actions.add(new OpenAction(ActivityField.MIMETYPE, this));
        for (Action action : Utilities.actionsForPath("Loaders/application/x-agrosense-activityfield/Actions")) {
            actions.add(action);
        }
        return actions.toArray(new Action[actions.size()]);
    }

    @Override
    public boolean canRename() {
        return true;
    }

    @Override
    public void setName(String s) {
        super.setName(s);
        activityField.setName(s);
    }

    /**
     * The Node has the name of object it represents
     * @return 
     */
    @Override
    public String getDisplayName() {
        return activityField.getName();
    }

    @Override
    public Image getIcon(int type) {
        return icon;
    }

    @Override
    public Image getOpenedIcon(int type) {
        return getIcon(type);
    }

    @Override
    public PasteType getDropType(Transferable t, int action, int index) {
        final Node node = NodeTransfer.node(t, NodeTransfer.DND_COPY);
        if (acceptNode(node)) {
            return new PasteType() {
                @Override
                public Transferable paste() throws IOException {

                    List<Shape> shapes = node.getLookup().lookup(ShapeCollection.class).getShapes();
                    for (Shape shape : shapes) {
                        if (!treatmentZoneWithGeometryExists(shape.getGeometry()) && activityField.geometryFits(shape.getGeometry())) {
                            activityField.addTreatmentZone(createTreatmentZoneFromShape(shape));
                        }
                    }
                    return null;
                }
            };
        } else {
            return null;
        }
    }

    /**
     * This method verifies if a Node can be used by the ActivityField
     * represented by this ActivityFieldNode. Nodes dropped on this
     * ActivityFieldNode can be used to add treatment zones if they meet the
     * right criteria tested by this method. The Nodes lookup needs to contains
     * a {@link ShapeCollection} object and the shapes contained by the
     * ShapeCollection may no be present in the ActivityField represented by
     * this ActivityFieldNode.
     *
     * @param node The Node that should be tested for usefulness
     * @return True if the Lookup of the passed Node contains a ShapeCollection
     * and the shapes of that collection are not present in the ActivityField of
     * this ActivityFieldNode, false otherwise
     */
    protected boolean acceptNode(Node node) {
        if (node != null) {
            ShapeCollection shapeCollection = node.getLookup().lookup(ShapeCollection.class);
            if (shapeCollection != null) {
                return true;
            }
        }
        return false;
    }

    private TreatmentZone createTreatmentZoneFromShape(Shape shape) {
        TreatmentZone treatmentZone = new TreatmentZone(activityField.getURI(), shape.getGeometry());
        treatmentZone.setName(shape.getName());
        return treatmentZone;
    }

    /**
     * This method checks if an ActivityField contains a Treatment Zone with a
     * specific geometry.
     *
     * @param geometry The geometry you wish to check the presence of
     * @return true of the geometry is contained by a treatment zone in the
     * ActivityField represented by this node, false otherwise
     */
    private boolean treatmentZoneWithGeometryExists(Geometry geometry) {
        List<TreatmentZone> existingTreatmentZones = activityField.getTreatmentZones();
        for (TreatmentZone treatmentZone : existingTreatmentZones) {
            if (treatmentZone.getGeometry().equals(geometry)) {
                return true;
            }
        }
        return false;
    }
}
