/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.field.type;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryCollection;
import com.vividsolutions.jts.geom.GeometryFactory;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;
import nl.cloudfarming.client.farm.model.FarmManager;
import nl.cloudfarming.client.field.model.*;
import nl.cloudfarming.client.model.*;
import org.jdesktop.swingworker.SwingWorker;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.netbeans.api.project.Project;
import org.openide.filesystems.*;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;
import org.openide.util.lookup.ServiceProviders;

/**
 *
 * @author Timon Veenstra
 */
@ServiceProviders({
    @ServiceProvider(service = FieldManager.class),
    @ServiceProvider(service = EntityManager.class)
})
@NbBundle.Messages({"FieldManagerImpl_progress_handle_display_name=Importing fields..."})
public class FieldManagerImpl extends AbstractEntityManager<Field> implements FieldManager {

    public static final String FIELDS_DIR = "Fields";
    private static final Set<ItemIdType> SUPPORTED_TYPES = EnumSet.of(ItemIdType.FLD, ItemIdType.AFD, ItemIdType.TZN);
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.field.type.FieldManagerImpl");

    @Override
    public Set<ItemIdType> getSupportedTypes() {
        return SUPPORTED_TYPES;
    }

    Field loadField(FileObject fileObject) {
        LOGGER.log(Level.FINEST, "loading Field object from file {0}", fileObject.getName());
        assert Field.EXT.equals(fileObject.getExt()) : "Can only load Fields";
        Object object = null;

        try (ObjectInputStream ois = new ObjectInputStream(fileObject.getInputStream())) {
            object = ois.readObject();
        } catch (ClassNotFoundException | IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not load file {0} with the exception {1}", new Object[]{fileObject, ex});
        }
        assert object instanceof Field : "fileobject did not contain the right type (Field)";
        Field field = (Field) object;

        LOGGER.log(Level.FINEST, "loaded field with {0} activity fields", field.getActivityFields().size());

        return getCachedInstanceFor(field);
    }

    @Override
    public void save(Field field) {
        FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
        Project project = farmManager.findFarmProject(field.getURI());
        save(field, project);
    }

    @Override
    public void remove(Field entity) {
        //FIXME implement remove field
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private void save(Object object, FileObject fileObject) {
        LOGGER.log(Level.FINEST, "saving field object to file {0}", fileObject.getName());
        assert Field.EXT.equals(fileObject.getExt()) : "Can only save Fields";
        // Try-with-resource seems to have slower performance
        try (ObjectOutputStream oos = new ObjectOutputStream(fileObject.getOutputStream())) {
            oos.writeObject(object);
        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not save file {0} with the exception {1}", new Object[]{fileObject, ex});
        }
    }
    
    private void save(Field field, OutputStream outputStream, String fileName) {
        try (ObjectOutputStream oos = new ObjectOutputStream(outputStream)) {
            oos.writeObject(field);
            addToCache(field);
        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not save file {0} with the exception {1}", new Object[]{fileName, ex});
        }        
    }

    private FileObject save(Field field, Project project) {
        LOGGER.log(Level.FINEST, "Saving field with id {0}", new Object[]{field.getItemIdNumber()});

        FileObject fieldsFolder = getFieldsFolder(project, true);
        FileObject fieldObject = fieldsFolder.getFileObject(field.getItemIdNumber(), Field.EXT);

        try {
            if (fieldObject == null) {
                String fileName = field.getItemIdNumber() + "." + Field.EXT;
                // reading by explorer sometimes causes race conditions when using createData():
                // "FSException: Cannot get exclusive access to .../path/to/file.ext (probably opened for reading)."
                try (OutputStream os = fieldsFolder.createAndOpen(fileName)) {
                    save(field, os, fileName);
                }
                fieldObject = fieldsFolder.getFileObject(fileName);
            }
            else {
                save(field, fieldObject);
            }

        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Could not create field file {0} with the exception {1}", new Object[]{field.getItemIdNumber(), ex});
        }
        return fieldObject;
    }

    /**
     * Get the Fields root Folder
     *
     * @param project the project to find the folder in
     * @param create if the folder should be created if not found
     * @return FileObject if found else null
     */
    public static FileObject getFieldsFolder(final Project project, boolean create) {
        //Get the stock directory, creating if deleted
        FileObject farmFolder = project.getProjectDirectory();
        FileObject fieldsFolder = farmFolder.getFileObject(FIELDS_DIR);

        // Create field folder if it doesn't exist
        if (fieldsFolder == null && create) {
            try {
                fieldsFolder = farmFolder.createFolder(FIELDS_DIR);
            } catch (IOException ex) {
                LOGGER.log(Level.SEVERE, "Could not create the folder {0} in folder {1}", new Object[]{FIELDS_DIR, farmFolder.getPath()});
            }
        }
        return fieldsFolder;
    }

    /**
     * Read and return all fields from this projects directory Will first find
     * the root folder via getFieldsSubFolder() and loop through it's child the
     * final path is project
     *
     * @return List of fields
     */
    @Override
    public List<Field> getFields(Project project) {
        List<Field> shapes = new ArrayList<>();
        FileObject fieldsDir = getFieldsFolder(project, true);

        for (FileObject fieldFile : fieldsDir.getChildren()) {
            try {
                DataObject dataObject = DataObject.find(fieldFile);
                if (dataObject != null) {
                    Field field = dataObject.getLookup().lookup(Field.class);
                    if (field != null) {
                        shapes.add(field);
                        addToCache(field);
                    }
                }
            } catch (DataObjectNotFoundException ex) {
                LOGGER.log(Level.SEVERE, "Could not get DataObject for FileObject {0}", ex.getMessage());
            }
        }

        return shapes;
    }

    /**
     *
     * {
     *
     * @inheritDoc }
     */
    @Override
    public List<Field> importTreatmentZones(List<Shape> shapes, Project project, String fieldName, PropertyChangeListener progressListener) {
        assert !SwingUtilities.isEventDispatchThread() : "This method is blocking and should not be executed from the AWT";
        ImportTreatmentZonesWorker worker = new ImportTreatmentZonesWorker(shapes, project, fieldName);
        worker.addPropertyChangeListener(progressListener);
        worker.execute();
        try {
            return worker.get();
        } catch (InterruptedException | ExecutionException ex) {
            Exceptions.printStackTrace(ex);
            return null;
        }
    }

    @Override
    public List<Field> importFields(List<Shape> shapes, Project project, PropertyChangeListener progressListener) {
        assert !SwingUtilities.isEventDispatchThread() : "This method is blocking and should not be executed from the AWT";
        ImportFieldsWorker worker = new ImportFieldsWorker(shapes, project);
        worker.addPropertyChangeListener(progressListener);
        worker.execute();
        try {
            return worker.get();
        } catch (InterruptedException | ExecutionException ex) {
            Exceptions.printStackTrace(ex);
            return null;
        }
    }


    /**
     *
     * @param shape
     * @param project
     * @param field
     */
    @Override
    public Field importTreatmentZone(Shape shape, Project project, Field target, String groupName) throws IOException {
        return importTreatmentZone(shape, project, target, groupName, true);
    }
    
    public Field importTreatmentZone(Shape shape, Project project, Field target, String activityFieldName, boolean autoSave) throws IOException {
        EnterpriseIdProvider eip = project.getLookup().lookup(EnterpriseIdProvider.class);

        Field field;
        if (target == null) {
            LOGGER.finest("Creating new field to add imported treatmentzone to");
            field = new Field(eip, shape.getName());
        } else {
            field = target;
        }

        ActivityField afd = getActivityField(field, activityFieldName);

        if (afd == null) {
            LOGGER.finest("Creating new activity field to add imported treatmentzone to");
            afd = new ActivityField(eip);
            afd.setName(activityFieldName);
            field.addActivityField(afd);
        }

        TreatmentZone treatMentZone = new TreatmentZone(eip);
        treatMentZone.setGeometry(shape.getGeometry());
        treatMentZone.setName(shape.getName());
        treatMentZone.setValue(shape.getValue());

        afd.addTreatmentZone(treatMentZone);

        if (autoSave) {
            save(field, project);
        }

        return field;
    }

    /**
     * Import the supplied shape into the context of the provided project
     *
     * @param shape
     * @param project
     */
    @Override
    public Field importField(Shape shape, Project project) throws IOException {
        EnterpriseIdProvider eip = project.getLookup().lookup(EnterpriseIdProvider.class);
        if (!projectContainsGeometryOfShape(project, shape)) {
            Field field = new Field(eip, shape.getName());
            field.setGeometry(shape.getGeometry());

            ActivityField group = new ActivityField(eip);
            group.setName(field.getName());

            TreatmentZone treatmentZone = new TreatmentZone(eip, shape.getGeometry());
            treatmentZone.setName(shape.getName());
            treatmentZone.setValue(shape.getValue());

            group.addTreatmentZone(treatmentZone);
            field.addActivityField(group);

            save(field, project);
            return field;
        }
        return null;
    }

    @Override
    protected Field load(AgroURI uri, Project project) {

        switch (uri.getItemIdType()) {
            case FLD:
                FileObject fieldObject = getFieldsFolder(project, true).getFileObject(uri.getItemIdNumber(), Field.EXT);
                return loadField(fieldObject);
            case AFD:
            case TZN: // falling through
                // TODO: load all fields and search for the requested AFD/TZN
                LOGGER.log(Level.WARNING, "entity with URI {0} should already have been available via loading the parent field", new Object[]{uri});
                return null;
            default:
                LOGGER.log(Level.WARNING, "cannot load unsupported entity type {0}", new Object[]{uri.getItemIdType()});
                return null;
        }
    }

    @Override
    protected void addToCache(Entity entity) {
        super.addToCache(entity);
        // cascade child entities:
        if (entity instanceof Field) {
            for (ActivityField afd : ((Field) entity).getActivityFields()) {
                addToCache(afd);
            }
        }
        if (entity instanceof ActivityField) {
            for (TreatmentZone tzn : ((ActivityField) entity).getTreatmentZones()) {
                addToCache(tzn);
            }
        }
    }

    @Override
    protected void removeFromCache(Entity entity) {
        super.removeFromCache(entity);
        // cascade child entities:
        if (entity instanceof Field) {
            for (ActivityField afd : ((Field) entity).getActivityFields()) {
                removeFromCache(afd);
            }
        }
        if (entity instanceof ActivityField) {
            for (TreatmentZone tzn : ((ActivityField) entity).getTreatmentZones()) {
                removeFromCache(tzn);
            }
        }
    }

    /**
     * This method iterates through the fields of a project to see if the
     * project contains a field with the geometry of the given shape.
     *
     * @param project The project that should be searched
     * @param shape The shape that the method needs to look for
     * @return True if the given shape is found in one of the fields or false if
     * the shape is not found.
     */
    private boolean projectContainsGeometryOfShape(Project project, Shape shape) {
        List<Field> fields = getFields(project);
        for (Field field : fields) {
            Geometry geometry = field.getGeometry();
            //if the shape has no geometry there is no harm in importing it
            if (geometry != null) {
                if (field.getGeometry().equals(shape.getGeometry())) {
                    return true;
                }
            }
        }
        return false;
    }
    
    /**
     * Find an activity field within a field by name.
     * 
     * @param field
     * @param activityFieldName
     * @return 
     */
    private ActivityField getActivityField(Field field, String activityFieldName) {
        ActivityField result = null;
        
        for (ActivityField afd : field.getActivityFields()) {
            System.out.println("Found activityField named: " + afd.getName() + " in Field: " + field.getName());
            if (activityFieldName.equals(afd.getName())) {
                LOGGER.log(Level.FINEST, "found existing group with name {0}", activityFieldName);
                result = afd;
                break;
            }
        }
        return result;
    }
    
    /**
     * Create a geometry for a field by joining all treatmentzone geometries
     * in the specified activity field.
     * 
     * @param field
     * @param activityFieldName 
     */
    private void createFieldGeometry(Field field, String activityFieldName) {
        if (field.getGeometry() != null) {
            return;
        }
        
        ActivityField afd = getActivityField(field, activityFieldName);
        if (afd == null) {
            return;
        }
        
        List<TreatmentZone> tzns = afd.getTreatmentZones();
        if (tzns == null || tzns.isEmpty()) {
            return;
        }
        
        GeometryFactory factory = null;
        List<Geometry> geometries = new LinkedList<>();
        for (TreatmentZone tzn : tzns) {
            if (tzn.getGeometry() != null) {
                // use factory from first non-empty tzn geometry:
                if (geometries.isEmpty()) {
                    factory = tzn.getGeometry().getFactory();
                }
                geometries.add(tzn.getGeometry());
            }
        }
        
        if (geometries.isEmpty() || factory == null) {
            return;
        }
        
        GeometryCollection geometryCollection = (GeometryCollection) factory.buildGeometry(geometries);
        Geometry result = geometryCollection.union();
        LOGGER.log(Level.FINE, "Created field geometry of type: {0}", result.getGeometryType());
        field.setGeometry(result);
    }

    private final class ImportTreatmentZonesWorker extends SwingWorker<List<Field>, Object> {

        private final List<Shape> shapes;
        private final Project project;
        private final Field field;

        public ImportTreatmentZonesWorker(List<Shape> shapes, Project project, String fieldName) {
            this.shapes = shapes;
            this.project = project;
            EnterpriseIdProvider eip = project.getLookup().lookup(EnterpriseIdProvider.class);
            this.field = new Field(eip, fieldName);
        }

        @Override
        protected List<Field> doInBackground() throws Exception {
            List<Field> result = new ArrayList<>();
            int progress = 0;
            String activityFieldName = field.getName();
            result.add(field);

            for (Shape shape : shapes) {
                try {
                    importTreatmentZone(shape, project, field, activityFieldName, false);
                } catch (IOException ex) {
                    LOGGER.log(Level.SEVERE, "Import of shape failed with exception: {0}", ex);
                }
                progress++;
                firePropertyChange(PROP_PROGRESS, progress - 1, progress);
            }
            
            if (progress > 0) {
                createFieldGeometry(field, activityFieldName);
            }
            
            save(field, project);

            firePropertyChange(PROP_FINISH, null, progress);
            return result;
        }
    }

    private final class ImportFieldsWorker extends SwingWorker<List<Field>, Object> {

        private final List<Shape> shapes;
        private final Project project;

        public ImportFieldsWorker(List<Shape> shapes, Project project) {
            this.shapes = shapes;
            this.project = project;
        }

        @Override
        protected List<Field> doInBackground() throws Exception {
            List<Field> result = new ArrayList<>();
            int progress = 0;
            for (Shape shape : shapes) {
                try {
                    result.add(importField(shape, project));
                } catch (IOException ex) {
                    LOGGER.log(Level.SEVERE, "Import of shape failed with exception: {0}", ex);
                }
                progress++;
                firePropertyChange(PROP_PROGRESS, progress - 1, progress);
            }
            firePropertyChange(PROP_FINISH, null, progress);
            return result;
        }
    }

    /**
     * Thread to execute the saving
     */
    @NbBundle.Messages("ImportWorker title=Importing shapes")
    @Deprecated
    private final class ImportWorker2 implements Runnable {

        private final List<Shape> shapes;
        private final Project project;
        private final FieldManager manager;
        private ProgressHandle progressHandle;
        private final String processTitle = Bundle.ImportWorker_title();
        private int progress = 0;

        public ImportWorker2(List<Shape> shapes, Project project) {
            this.shapes = shapes;
            this.project = project;
            manager = Lookup.getDefault().lookup(FieldManager.class);
            assert manager != null : "Manager should not be null can not save shapes";
            progressHandle = ProgressHandleFactory.createSystemHandle(processTitle);
        }

        @Override
        public void run() {
            progressHandle.start(shapes.size());
            Field treatmentZoneField = null;
            for (Shape shape : shapes) {
                try {
                    switch (shape.getImportType()) {
                        case TREATMENT_ZONE:
                            treatmentZoneField = manager.importTreatmentZone(shape, project, treatmentZoneField, "import");
                            break;
                        case FIELD:
                            manager.importField(shape, project);
                            break;
                    }
                    progressHandle.progress(progress++);
                } catch (IOException ex) {
                    LOGGER.log(Level.SEVERE, "Import of shape failed with exception: {0}", ex);
                }
            }
            progressHandle.finish();
        }
    }
}
